// Device-specific include file.
#include <p33fxxxx.h>

// Standard C library headers.
#include <math.h>
#include <stdlib.h>
#include <string.h>

// Utility functions.
#include <libpic30.h>

#include "vdc.h"
#include "vdc_graphics.h"
#include "vdc_text.h"
#include "vdc_text_font.h"

// Configuration.

_FOSC(OSCIOFNC_ON & POSCMD_XT)
_FOSCSEL(FNOSC_PRIPLL)
_FWDT(FWDTEN_OFF)

// Constants.
#define M_PI (3.1415926535897932384626433832795f)

// Types.
struct point_3d_t {
	int x;
	int y;
	int z;
};

// Draws a swirly line pattern.
void demo_swirl(void) {

	float angle, sub_angle;
	unsigned int amplitude;
	unsigned int pass;

	struct vdc_point_2d_t a;
	struct vdc_point_2d_t b;

	vdc_graphics_set_colour(0);
	vdc_graphics_clear();
	vdc_graphics_set_colour(127);

	for (pass = 0; pass < 2; ++pass) {
		// Draw the pattern.		
		for (amplitude = 0, angle = 0; amplitude < 320; amplitude += 4, angle += 0.08f) {
			for (sub_angle = 0; sub_angle < 2 * M_PI; sub_angle += M_PI * 0.5f) {
				a.x = 160 + amplitude * sin(angle + sub_angle);
				a.y = 120 + amplitude * cos(angle + sub_angle);
				b.x = 160 + amplitude * sin(angle + sub_angle + M_PI * 0.5f);
				b.y = 120 + amplitude * cos(angle + sub_angle + M_PI * 0.5f);
				vdc_graphics_draw_line(a, b);
			}
		}
		if (pass == 0) {
			// Delay for two seconds.
			vdc_delay_frames(vdc_fps * 2);
			// Next loop, we'll clear the screen.
			vdc_graphics_set_colour(0);
		}
	}

	// Put the plotting mode back the way it was.
	vdc_graphics_set_colour(127);
}

// Face indices for the cube.
#define CUBE_FACE_FRONT 0
#define CUBE_FACE_BACK 1
#define CUBE_FACE_TOP 2
#define CUBE_FACE_BOTTOM 3
#define CUBE_FACE_LEFT 4
#define CUBE_FACE_RIGHT 5

// Vertex indices for the cube.
#define CUBE_FACE_FRONT_VA 3
#define CUBE_FACE_FRONT_VB 7
#define CUBE_FACE_FRONT_VC 5
#define CUBE_FACE_FRONT_VD 1

#define CUBE_FACE_BACK_VA 6
#define CUBE_FACE_BACK_VB 2
#define CUBE_FACE_BACK_VC 0
#define CUBE_FACE_BACK_VD 4

#define CUBE_FACE_TOP_VA 2
#define CUBE_FACE_TOP_VB 6
#define CUBE_FACE_TOP_VC 7
#define CUBE_FACE_TOP_VD 3

#define CUBE_FACE_BOTTOM_VA 0
#define CUBE_FACE_BOTTOM_VB 1
#define CUBE_FACE_BOTTOM_VC 5
#define CUBE_FACE_BOTTOM_VD 4

#define CUBE_FACE_LEFT_VA 2
#define CUBE_FACE_LEFT_VB 3
#define CUBE_FACE_LEFT_VC 1
#define CUBE_FACE_LEFT_VD 0

#define CUBE_FACE_RIGHT_VA 7
#define CUBE_FACE_RIGHT_VB 6
#define CUBE_FACE_RIGHT_VC 4
#define CUBE_FACE_RIGHT_VD 5

int triangle_is_clockwise(struct vdc_point_2d_t a, struct vdc_point_2d_t b, struct vdc_point_2d_t c) {
	return (((long)a.x * (long)b.y - (long)b.x * (long)a.y + (long)b.x * (long)c.y - (long)c.x * (long)b.y + (long)c.x * (long)a.y - (long)a.x * (long)c.y) >= 0);
}

void fill_quad(struct vdc_point_2d_t a, struct vdc_point_2d_t b, struct vdc_point_2d_t c, struct vdc_point_2d_t d) {
	vdc_graphics_fill_triangle(a, b, c);
	vdc_graphics_fill_triangle(c, d, a);
}

int shade_face(float x, float y, float z) {
	// Calculate the clour 
	int colour = 16 + (120.0 * ((x * -0.25) + (y * -0.35) + (z * -0.9)));
	if (colour < 0) {
		return 0;
	} else if (colour > 127) {
		return 127;
	} else {
		return colour;
	}
}

// Draws a spinning cube.
void demo_cube(void) {
	

	// Source X, Y and Z of each point2d of the cube.
	int x, y, z;

	// Rotation in X and Y.
	float rx = 0, ry = 0;
	
	// Computed sin/cos of the rotated X and Y.
	float sin_rx, cos_rx, sin_ry, cos_ry;
	
	// Transformed X, Y and Z.
	float tx, ty, tz;

	// Stores the eight projected vertices.
	struct vdc_point_2d_t vertices[8];

	// Determines which of the faces to draw (that is, which have not been culled).
	unsigned int faces_to_draw;

	// Stores the computed colour of each face.
	unsigned int face_colours[6];

	// There's always a need for one of these.
	unsigned int i;

	// Size of the cube.
	float rs = 0;
	float s;
	
	int demo_timer = 315 * 4;
	
	while (demo_timer-- > 0) {

		// Calculate the sin/cos of the rotated X and Y.
		sin_rx = sin(rx);
		cos_rx = cos(rx);
		sin_ry = sin(ry);
		cos_ry = cos(ry);

		// Calculate the size of the cube.
		if (vdc_output_mode == VDC_OUTPUT_VGA) {
			s = 70.0f * (1.0f - cos(rs));
		} else {
			s = 100.0f * (1.0f - cos(rs));
		}
	
		// Calculate each vertex in turn.
		i = 0;
		for (x = -1; x <= +1; x += 2) {
			for (y = -1; y <= +1; y += 2) {
				for (z = -1; z <= +1; z+= 2) {
	
					// Calculate the transformed vertex.
					tx = y * cos_rx - x * sin_rx;
					ty = -x * cos_rx * sin_ry - y * sin_rx * sin_ry - z * cos_ry;
	  				tz = 3 - x * cos_rx * cos_ry - y * sin_rx * cos_ry + z * sin_ry;
	
					// Project and store.
					vertices[i].x = 160 + (int)(tx * s / tz);
					vertices[i].y = 120 + (int)(ty * s / tz);
					
					// Advance to the next vertex.
					++i;
				}
			}
		}
		
		// Clear which faces to draw.
		faces_to_draw = 0;

		// Front and back.
		if (triangle_is_clockwise(vertices[CUBE_FACE_FRONT_VA], vertices[CUBE_FACE_FRONT_VB], vertices[CUBE_FACE_FRONT_VC])) {
			face_colours[CUBE_FACE_FRONT] = shade_face(0, -cos_ry, sin_ry);
			faces_to_draw |= (1 << CUBE_FACE_FRONT);
		}
		if (triangle_is_clockwise(vertices[CUBE_FACE_BACK_VA], vertices[CUBE_FACE_BACK_VB], vertices[CUBE_FACE_BACK_VC])) {
			face_colours[CUBE_FACE_BACK] = shade_face(0, cos_ry, -sin_ry);
			faces_to_draw |= (1 << CUBE_FACE_BACK);
		}

		// Top and bottom.
		if (triangle_is_clockwise(vertices[CUBE_FACE_TOP_VA], vertices[CUBE_FACE_TOP_VB], vertices[CUBE_FACE_TOP_VC])) {
			face_colours[CUBE_FACE_TOP] = shade_face(cos_rx, -sin_rx * sin_ry, -sin_rx * cos_ry);
			faces_to_draw |= (1 << CUBE_FACE_TOP);
		}
		if (triangle_is_clockwise(vertices[CUBE_FACE_BOTTOM_VA], vertices[CUBE_FACE_BOTTOM_VB], vertices[CUBE_FACE_BOTTOM_VC])) {
			face_colours[CUBE_FACE_BOTTOM] = shade_face(-cos_rx, sin_rx * sin_ry, sin_rx * cos_ry);
			faces_to_draw |= (1 << CUBE_FACE_BOTTOM);
		}

		// Left and right.
		if (triangle_is_clockwise(vertices[CUBE_FACE_LEFT_VA], vertices[CUBE_FACE_LEFT_VB], vertices[CUBE_FACE_LEFT_VC])) {
			face_colours[CUBE_FACE_LEFT] = shade_face(sin_rx, cos_rx * sin_ry, cos_rx * cos_ry);
			faces_to_draw |= (1 << CUBE_FACE_LEFT);
		}
		if (triangle_is_clockwise(vertices[CUBE_FACE_RIGHT_VA], vertices[CUBE_FACE_RIGHT_VB], vertices[CUBE_FACE_RIGHT_VC])) {
			face_colours[CUBE_FACE_RIGHT] = shade_face(-sin_rx, -cos_rx * sin_ry, -cos_rx * cos_ry);
			faces_to_draw |= (1 << CUBE_FACE_RIGHT);
		}
		
		// Wait for a vblank before starting.
		vdc_wait_vblank();

		// Clear the entire screen (fast!)
		memset(vdc_frame_buffer, 0, sizeof(vdc_frame_buffer));

		// Front and back.
		vdc_graphics_set_colour(32);
	 	if (faces_to_draw & (1 << CUBE_FACE_FRONT)) {
			vdc_graphics_set_colour(face_colours[CUBE_FACE_FRONT]);
			fill_quad(vertices[CUBE_FACE_FRONT_VA], vertices[CUBE_FACE_FRONT_VB], vertices[CUBE_FACE_FRONT_VC], vertices[CUBE_FACE_FRONT_VD]);
		}
	 	if (faces_to_draw & (1 << CUBE_FACE_BACK)) {
			vdc_graphics_set_colour(face_colours[CUBE_FACE_BACK]);
			fill_quad(vertices[CUBE_FACE_BACK_VA], vertices[CUBE_FACE_BACK_VB], vertices[CUBE_FACE_BACK_VC], vertices[CUBE_FACE_BACK_VD]);
		}

		// Top and bottom.
	 	if (faces_to_draw & (1 << CUBE_FACE_TOP)) {
			vdc_graphics_set_colour(face_colours[CUBE_FACE_TOP]);
			fill_quad(vertices[CUBE_FACE_TOP_VA], vertices[CUBE_FACE_TOP_VB], vertices[CUBE_FACE_TOP_VC], vertices[CUBE_FACE_TOP_VD]);
		}
	 	if (faces_to_draw & (1 << CUBE_FACE_BOTTOM)) {
			vdc_graphics_set_colour(face_colours[CUBE_FACE_BOTTOM]);
			fill_quad(vertices[CUBE_FACE_BOTTOM_VA], vertices[CUBE_FACE_BOTTOM_VB], vertices[CUBE_FACE_BOTTOM_VC], vertices[CUBE_FACE_BOTTOM_VD]);
		}

		// Left and right.
	 	if (faces_to_draw & (1 << CUBE_FACE_LEFT)) {
			vdc_graphics_set_colour(face_colours[CUBE_FACE_LEFT]);
			fill_quad(vertices[CUBE_FACE_LEFT_VA], vertices[CUBE_FACE_LEFT_VB], vertices[CUBE_FACE_LEFT_VC], vertices[CUBE_FACE_LEFT_VD]);
		}
	 	if (faces_to_draw & (1 << CUBE_FACE_RIGHT)) {
			vdc_graphics_set_colour(face_colours[CUBE_FACE_RIGHT]);
			fill_quad(vertices[CUBE_FACE_RIGHT_VA], vertices[CUBE_FACE_RIGHT_VB], vertices[CUBE_FACE_RIGHT_VC], vertices[CUBE_FACE_RIGHT_VD]);
		}

		switch (vdc_output_mode) {
			case VDC_OUTPUT_TV:
			case VDC_OUTPUT_VGA:
				rs += 0.02f / 4.0f;
				ry += 0.03f / 4.0f;
				rx += 0.05f / 4.0f;
				break;
			case VDC_OUTPUT_LCD:
				rs += 0.02f * 2.0f;
				ry += 0.03f * 2.0f;
				rx += 0.05f * 2.0f;
				vdc_delay_frames(7);
				demo_timer -= 7;
				break;
		}
	}
	vdc_graphics_set_colour(0);
	vdc_graphics_clear();
	vdc_graphics_set_colour(127);
}

// Draws a Sierpinksi triangle.
void demo_sierpinski(void) {
	
	float x, y, n;
	unsigned int steps = 15000;

	vdc_graphics_set_colour(0);
	vdc_graphics_clear();
	vdc_graphics_set_colour(127);
	
	srand(0);

	x = rand() / (float)RAND_MAX;
	y = rand() / (float)RAND_MAX;
	
	while (steps--) {
		n = rand() / (float)RAND_MAX;
		if (n <= 1.0f / 3.0f) {
			x *= 0.5f;
			y *= 0.5f;
		} else if (n <= 2.0f / 3.0f) {
			x = (0.5f + x) * 0.5f;
			y = (1.0f + y) * 0.5f;
		} else {
			x = (1.0f + x) * 0.5f;
			y *= 0.5f;
		}
		vdc_graphics_set_pixel(x * 320, (1.0f - y) * 240);
	}

	vdc_delay_frames(vdc_fps);
}

// Draws a Mandelbrot fractal. Based on code by Ken Silverman.
void demo_mandelbrot(void) {

	// Device (x,y) coordinates.
	int x, y;
	
	// Used to calculate the colour at each point2d.
	float mx, ox, my, oy, xx, yy;

	float hundredth = 1.0f / 100.0f;
	
	// Stores the working colour.
	unsigned int c;

	// Used to dither the results. Taken from http://en.wikipedia.org/wiki/Ordered_dithering
	unsigned int threshold_map[4][4] = {
		{  1,  9,  3, 11 },
		{ 13,  5, 15,  7 },
		{  4, 12,  2, 10 },
		{ 16,  8, 14,  6 }
	};

	for (y = 0; y < 240; ++y) {
		vdc_auto_init_output();
		for (x = 0; x < 320; ++x) {
	
			// Calculate coordinates of point2d to sample in the fractal.
			ox = mx = (x - 160) * hundredth - 0.7f;
			oy = my = (y - 120) * hundredth;

			// Calculate the colour at the point2d.
			for (c = 0; c < 16; ++c) {
				xx = mx * mx;
				yy = my * my;
				if (xx + yy >= 4) break;
				my = mx * my * 2.0f + oy;
				mx = xx - yy + ox;
			}
			
			// Colour the pixel.
			if (c + threshold_map[x & 3][y & 3] > 16) {
				vdc_graphics_set_pixel(x, y);
			}
		}
	}
}

// Flips the screen upside down four times.
void demo_flip(void) {

	unsigned int top_row, bottom_row, column;
	unsigned int* top_ptr;
	unsigned int* bottom_ptr;
	unsigned int flip_count;
	
	for (flip_count = 0; flip_count < 4; ++flip_count) {
		vdc_wait_vblank();
		for (top_row = 0, bottom_row = 239; top_row < 120; ++top_row, --bottom_row) {
			top_ptr = vdc_frame_buffer + 20 * top_row;
			bottom_ptr = vdc_frame_buffer + 20 * bottom_row;
			for (column = 0; column < 20; ++column) {
				unsigned int s = *top_ptr;
				*(top_ptr++) = *bottom_ptr;
				*(bottom_ptr++) = s;
			}
		}
		vdc_delay_frames(vdc_fps / 2);
	}

}

// Draws a circular wipe pattern on the screen.
void demo_circle_wipe(void) {
	unsigned int r, jiggle;
	struct vdc_point_2d_t origin = { 160, 120 };
	for (r = 0; r <= 220; ++r) {
		vdc_wait_vblank();
		for (jiggle = 0; jiggle < 2; ++jiggle) {
			vdc_graphics_draw_circle(origin, r);
			++origin.x;
		}
		origin.x -= 2;
		if (r > 20) {
			vdc_graphics_set_colour(0);
			for (jiggle = 0; jiggle < 2; ++jiggle) {
				vdc_graphics_draw_circle(origin, r - 20);
				++origin.x;
			}
			origin.x -= 2;
			vdc_graphics_set_colour(127);
		}
	}
}

// Draw a clock hand at a specified angle and with a specified length.
void demo_clock_draw_hand(float angle, int length) {

	struct vdc_point_2d_t a = { 160 + length * sin(angle), 120 - length * cos(angle) }, b = { 160, 120 };

	int arrowhead_angle_offset;

	// Main part of the hand.
	vdc_graphics_draw_line(a, b);
	
	// Draw the arrowhead.
	for (arrowhead_angle_offset = -1; arrowhead_angle_offset <= +1; arrowhead_angle_offset += 2) {
		b.x = 160 + length * 0.85f * sin(angle + arrowhead_angle_offset * 0.1f);
		b.y = 120 - length * 0.85f * cos(angle + arrowhead_angle_offset * 0.1f);
		vdc_graphics_draw_line(a, b);
	}
}

// Draws the clock demo.
void demo_clock(void) {

	int jitter;
	int hand_position;

	vdc_graphics_set_colour(0);
	vdc_graphics_clear();

	vdc_graphics_set_colour(127);

	struct vdc_point_2d_t a = { 160, 120 }, b = { 160, 120 };

	// Draw the outer dial.
	--a.x; --b.y;
	for (jitter = -1; jitter <= 0; ++jitter) {
		// Outer ring.
		vdc_graphics_draw_circle(a, 118);
		vdc_graphics_draw_circle(b, 118);
		// Inner ring.
		vdc_graphics_draw_circle(a, 105);
		vdc_graphics_draw_circle(b, 105);
		// Offset for jitter.
		++a.x; ++b.y;
	}
	--a.x; --b.y;
	
	// Draw the marks around the dial.
	for (hand_position = 0; hand_position < 60; ++hand_position) {
		float tick_length = (hand_position % 5) ? 97 : 90;
		float angle = (hand_position / 30.0f) * M_PI;
		a.x = 160 + 100 * sin(angle);
		a.y = 120 - 100 * cos(angle);
		b.x = 160 + tick_length * sin(angle);
		b.y = 120 - tick_length * cos(angle);
		vdc_graphics_draw_line(a, b);
	}

	for (hand_position = 55; hand_position <= 60; ++hand_position) {
		
		// Draw the current hand position.
		demo_clock_draw_hand((hand_position / 30.0f) * M_PI, 88);
		// Delay one second.
		vdc_delay_frames(vdc_fps);

		// Have we finished?
		if (hand_position == 60) break;

		// Erase the old hand before drawing the next one.
		vdc_graphics_set_colour(0);
		demo_clock_draw_hand((hand_position / 30.0f) * M_PI, 88);
		vdc_graphics_set_colour(127);
	}
}

// Slides the image off the bottom of the LCD.
void demo_slide(void) {
	int row;
	unsigned char* vdc_frame_buffer_byte = (unsigned char*)vdc_frame_buffer;
	for (row = 0; row < 240; row += 2) {
		vdc_wait_vblank();
		memmove(vdc_frame_buffer_byte + (row + 2) * 40, vdc_frame_buffer_byte + row * 40, 9600 - (row + 2) * 40);
		memset(vdc_frame_buffer_byte + row * 40, 0, 40 * 2);
	}
}

// Total number of stars = sizeof(vdc_text_buffer)/sizeof(point_3d_t)=(80*30)/(3*2)=2400/6=400.
#define STAR_COUNT 400

// Runs a starfield demo.
void demo_stars(void) {

	struct point_3d_t* stars = (struct point_3d_t*)vdc_text_buffer;

	int star;
	struct vdc_point_2d_t star_transformed;
	struct vdc_point_2d_t previous_star_transformed;
	int speed;

	for (star = 0; star < STAR_COUNT; ++star) stars[star].z = 1;

	vdc_graphics_set_colour(0);
	vdc_graphics_clear();

	int frames_run = 0;
	int stars_drawn = 0;
	int stars_to_draw = 0;

	do {

		// We haven't drawn any stars.
		stars_drawn = 0;

		// Wait for vblank or a particular number of frames.
		// The LCD requires slower, choppier animation due to its slow response.
		switch (vdc_output_mode) {
			case VDC_OUTPUT_TV:
			case VDC_OUTPUT_VGA:
				speed = 5;
				vdc_wait_vblank();
				++frames_run;
				break;
			case VDC_OUTPUT_LCD:
				speed = 20;
				vdc_delay_frames(4);
				frames_run += 4;
				break;
		}
	
		// Increase the number of stars to draw.
		if (stars_to_draw < STAR_COUNT) {
			stars_to_draw = frames_run;
			if (stars_to_draw > STAR_COUNT) stars_to_draw = STAR_COUNT;
		}

		// Move and draw each star.
		for (star = 0; star < stars_to_draw; ++star) {

			previous_star_transformed.x = stars[star].x / stars[star].z + 160;
			previous_star_transformed.y = stars[star].y / stars[star].z + 120;
			
			if ((stars[star].z -= speed) < 50 || previous_star_transformed.x < 0 || previous_star_transformed.x > 319 || previous_star_transformed.y < 0 || previous_star_transformed.y > 239) {
				vdc_graphics_clear_pixel(previous_star_transformed.x, previous_star_transformed.y);
				if (frames_run < 1000) {
					// We're still animating, so add a star.
					stars[star].x = ((rand() & 0x1FF) - 256) << 7;
					stars[star].y = ((rand() & 0x1FF) - 256) << 7;
					stars[star].z = 300 + (rand() & 511);
					++stars_drawn;
				} else {
					// Don't bother to redraw the star.
					stars[star].z = 1;
				}
			} else {
				star_transformed.x = stars[star].x / stars[star].z + 160;
				star_transformed.y = stars[star].y / stars[star].z + 120;
				vdc_graphics_clear_pixel(previous_star_transformed.x, previous_star_transformed.y);
				vdc_graphics_set_pixel(star_transformed.x, star_transformed.y);
				++stars_drawn;
			}
		}
	} while (stars_drawn > 0);
	
	// Ensure the LCD is clear.
	vdc_graphics_set_colour(0);
	vdc_graphics_clear();
	vdc_graphics_set_colour(127);
}

// Draws a single shaded ball at a specified position and size on the screen.
void draw_ball(struct vdc_point_2d_t origin, int size) {
	struct vdc_point_2d_t circle_origin;
	int shade, shade_size;
	for (shade = 15; shade >= 0; --shade) {
		vdc_graphics_set_colour_and_plot_mode((shade << 3) | (shade >> 1), VDC_GRAPHICS_PLOT_AND);
		circle_origin.x = origin.x - (shade * (160L - origin.x) * size) / 4000L;
		circle_origin.y = origin.y + size;
		shade_size = size * sin(shade / 32.0 * M_PI);
		vdc_graphics_fill_ellipse(circle_origin, (shade_size * 5) / 4, shade_size / 2);
	}
	for (shade = 0; shade < 16; ++shade) {
		vdc_graphics_set_colour((shade << 3) | (shade >> 1));
		circle_origin.x = origin.x + (shade * (160L - origin.x) * size) / 3000L;
		circle_origin.y = origin.y + (shade * (120L - origin.y) * size) / 3000L;
		vdc_graphics_fill_circle(circle_origin, size * cos(shade / 32.0 * M_PI));
	}
}

// Draws a field of shaded balls.
void demo_balls(void) {
	
	// Clear to black.
	vdc_graphics_set_colour(0);
	vdc_graphics_clear();
	
	// Fade in the ground.
	vdc_graphics_viewport.top = 70;
	int fade_in;
	for(fade_in = 0; fade_in < 16; ++fade_in) {
		vdc_graphics_set_colour((fade_in << 3) | (fade_in >> 1));
		vdc_graphics_clear();
		vdc_delay_frames(vdc_fps / 16);
	}
	vdc_graphics_viewport.top = 0;


	// Draw the field of balls.
	int x, y;
	struct vdc_point_2d_t ball_origin;
	int ball_size;
	for (y = 12; y >= 1; --y) {

		vdc_delay_frames(vdc_fps / 10);
		ball_origin.y = 240 / y + 50;
		ball_size = 75 / y;
		for (x = -12; x <= +12; ++x) {
			ball_origin.x = (x * 170) / y + 160;
			if ((ball_origin.x + ball_size) >= 0 && (ball_origin.x - ball_size) < 320) {
				draw_ball(ball_origin, ball_size);
			}
		}
	}
	vdc_delay_frames(vdc_fps * 2);
	
	// Fade out.
	int fade_out;
	for(fade_out = 15; fade_out >= 0; --fade_out) {
		vdc_delay_frames(vdc_fps / 8);
		vdc_graphics_set_colour_and_plot_mode((fade_out << 3) | (fade_out >> 1), VDC_GRAPHICS_PLOT_AND);
		vdc_graphics_clear();
	}
	// Reset to the default colour.
	vdc_graphics_set_colour(127);
}

const char* demo_title_messages = 
	"\n"
	"    dsPIC33 VDC\n"
	"\n"
	"320x240 resolution."
	"Outputs to a PAL TV"
	"or a graphical LCD."
	"\n"
	"  Ben Ryves   2010";

// Print a NUL-terminated message on the screen.
void print_message(const char* message) {
	char c;
	while ((c = *message++)) {
		switch (c) {
			case '\n':
				vdc_text_cursor_position.x = vdc_text_viewport.left;
				vdc_text_move_cursor_down();
				break;
			default:
				vdc_text_put_char(c);
				break;
		}
	}
}

// Show a title card.
void demo_title(void) {
	// Change to the large font.
	vdc_text_set_font_size(VDC_TEXT_FONT_SIZE_LARGE);
	// Fade in the background.
	int shade;
	for (shade = 0; shade < 16; ++shade) {
		vdc_delay_frames(vdc_fps / 16);
		vdc_graphics_set_colour((shade << 3) | (shade >> 1));
		vdc_graphics_clear();
	}
	// Draw the shadow.
	struct vdc_rectangle_t r = { 64, 64, 263, 183 };
	for (shade = 15; shade >= 0; --shade) {
		++r.left; ++r.top;
		--r.right; --r.bottom;
		vdc_graphics_set_colour_and_plot_mode((shade << 3) | (shade >> 1), VDC_GRAPHICS_PLOT_AND);
		vdc_graphics_fill_rectangle(r);
	}
	// Draw the outline of the rectangle.
	r.left -= 16; r.top -= 16;
	vdc_graphics_set_colour(127);
	vdc_graphics_fill_rectangle(r);
	vdc_graphics_set_colour(0);
	vdc_graphics_draw_rectangle(r);
	// Set the text viewport.
	r.left = (r.left / 8) + 2; r.top = (r.top / 8) + 2;
	r.right = (r.right / 8) - 2; r.bottom = (r.bottom / 8) - 2;
	vdc_text_set_viewport(r);
	// Set the text colour.
	vdc_text_foreground_colour = 0;
	vdc_text_background_colour = 127;
	// Draw the title card message.
	print_message(demo_title_messages);
	// Underline the title.
	struct vdc_point_2d_t a = { 113, 97 };
	struct vdc_point_2d_t b = { a.x + 11 * 8 - 2, a.y };
	vdc_graphics_draw_line(a, b);
	// Delay with the message on the screen.
	vdc_delay_frames(vdc_fps * 4);
	// Reset the text viewport and colours.
	vdc_text_reset_viewport();
	vdc_text_foreground_colour = 127;
	vdc_text_background_colour = 0;
}

// Wipe the screen using a grid of circles.
void demo_circle_grid_wipe(void) {
	struct vdc_point_2d_t o;
	int total_wipe_size = 0, current_circle_wipe_size;
	do {
		vdc_delay_frames(2);
		for (current_circle_wipe_size = total_wipe_size, o.y = -8; o.y <= 248; o.y += 32) {
			for (o.x = 16; o.x < 320; o.x += 32, current_circle_wipe_size -= 1) {
				if (current_circle_wipe_size >= 0 && current_circle_wipe_size < 32) {
					vdc_graphics_fill_circle(o, current_circle_wipe_size);
				}
			}
		}
		++total_wipe_size;
	} while (current_circle_wipe_size < 32);
}

int main(void) {

	// Configure PLL prescaler, PLL postscaler, PLL divisor
	
	// Fosc=Fin*(M/(N1*N2))
	PLLFBD = 30; // M = 32
	CLKDIVbits.PLLPOST = 0; // N1=2
	CLKDIVbits.PLLPRE = 0; // N2=2

	// Initialise the VDC.
	vdc_reset();
	vdc_auto_init_output();

	// Wait a little for the TV to initialise.
	vdc_delay_frames(vdc_fps);

	// Run the demos.
	demo_title();
	demo_circle_grid_wipe();
	demo_clock();
	demo_slide();
	demo_stars();
	demo_sierpinski();
	demo_flip();
	demo_circle_wipe();
	demo_balls();
	demo_swirl();
	demo_cube();
	demo_mandelbrot();
	for(;;) vdc_wait_vblank();

}
